/* 
** Copyright 1994 by Miron Livny, and Mike Litzkow
** 
** Permission to use, copy, modify, and distribute this software and its
** documentation for any purpose and without fee is hereby granted,
** provided that the above copyright notice appear in all copies and that
** both that copyright notice and this permission notice appear in
** supporting documentation, and that the names of the University of
** Wisconsin and the copyright holders not be used in advertising or
** publicity pertaining to distribution of the software without specific,
** written prior permission.  The University of Wisconsin and the 
** copyright holders make no representations about the suitability of this
** software for any purpose.  It is provided "as is" without express
** or implied warranty.
** 
** THE UNIVERSITY OF WISCONSIN AND THE COPYRIGHT HOLDERS DISCLAIM ALL
** WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES
** OF MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE UNIVERSITY OF
** WISCONSIN OR THE COPYRIGHT HOLDERS BE LIABLE FOR ANY SPECIAL, INDIRECT
** OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS
** OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE
** OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE
** OR PERFORMANCE OF THIS SOFTWARE.
** 
** Author:  Mike Litzkow
**
*/ 

/*
  This program exercises the condor file stream protocol.  This is intended
  to be a fast way to transfer an entire file either to or from memory.
  To use the protocol, you need to use either the CONDOR_put_file_stream,
  or CONDOR_get_file_stream RPC.  These result in an IP address and a
  port number to connect to.  Those can be turned into a file descriptor
  using the open_tcp_stream() routine, which is included here.  These
  transfers involve a length which is an input parameter to
  CONDOR_put_file_stream, and an output parameter from CONDOR_get_file_stream.
  The user code is expected to read or write the correct number of
  bytes, then close the file.  Since the shadow spawns a new process
  to handle the stream transfer, it is OK to mix other remote system
  calls with reads() and writes() on the streams.
*/

#define _POSIX_SOURCE

#include "condor_common.h"
#include "condor_syscall_mode.h"
#include "condor_debug.h"
#include "condor_constants.h"
#include "syscall_numbers.h"
#include <sys/socket.h>
#include <netinet/in.h>

static char *_FileName_ = __FILE__;

#define MIN(a,b) ((a)<(b)?(a):(b))
#define CHUNK_SIZE 4096

int open_tcp_stream( unsigned int ip_addr, unsigned short port );
char *read_file( const char *name, size_t *len );
void write_file( char *data, size_t len, const char *name );

char	Buf[] = "Hello World\n";

void
usage( const char *my_name )
{
	fprintf( stderr, "Usage: %s src dst\n", my_name );
	exit( 13 );
}

main( int argc, char *argv[] )
{
	int		i;
	char	*data;
	size_t	len;

	if( argc != 3 ) {
		usage( argv[0] );
	}

	data = read_file( argv[1], &len );
	write_file( data, len, argv[2] );
	exit( 0 );
}

char *
read_file( const char *name, size_t *len )
{
	int		status;
	int		fd;
	unsigned short	port;
	unsigned int	addr;
	size_t	nbytes, bytes_to_go;
	char	*data, *ptr;

	fd = open_file_stream( name, O_RDONLY, len );
	ASSERT( fd >= 0 );

		/* Allocate space for the file */
	data = malloc( *len );
	ASSERT( data );

		/* Read the whole file */
	ptr = data;
	for( bytes_to_go = *len; bytes_to_go; bytes_to_go -= nbytes ) {
		nbytes = MIN( CHUNK_SIZE, bytes_to_go );
		nbytes = read( fd, ptr, nbytes );
		ptr += nbytes;
		/*
		getuid();
		*/
	}

		/* clean up */
	close( fd );
	free( data );

		/* report */
	printf( "Read %d bytes\n", *len );
	return data;
}

void
write_file( char *data, size_t len, const char *name )
{
	int		status;
	int		fd;
	unsigned short	port;
	unsigned int	addr;
	size_t	nbytes, bytes_to_go;
	char	*ptr;
	int		scm;

	fd = open_file_stream( name, O_WRONLY, &len );
	ASSERT( fd >= 0 );
	
		/* Write the whole file */
	ptr = data;
	for( bytes_to_go = len; bytes_to_go; bytes_to_go -= nbytes ) {
		nbytes = MIN( CHUNK_SIZE, bytes_to_go );
		status = write( fd, ptr, nbytes );
		ASSERT( status == nbytes );
		ptr += nbytes;
	}

		/* clean up */
	close( fd );

		/* report */
	printf( "Wrote %d bytes\n", len );
}
