#!/usr/bin/perl -w

#
# This perl script generates an SVR4 "Node" file describing various tty device 
# files for the SAS driver.  
#
# Unless invoked with -d (use defaults), it interactively asks which kind of
# device files you want.
#
# Author: Jim Avera (jima@netcom.com) 8/31/93
# Public Domain.
#

# DEFAULTS if user just types return to prompts or if invoked with "-d" arg:
$nports  = 2;		# Number of ports
$flow    = "yes";	# Create devices with forced hardware flow-control?
$noflow  = "yes";	# Create regular devices (without hardware flow ctl)?
$direct  = "yes";	# Create direct (dumb 3-wire) devices?
$termdir = "yes";	# Create /dev/term/ aliases for modem ports?

# DEVICE FILE NAMING CONVENTIONS 
#  The following printf formats are used to print the device number along
#  with other characters indicating what kind of device file it is.
#
#  Un-comment the appropriate block, or add your own.
#
#  Variable naming convention: I=dial-in, O=dial-out, D=dumb(3 wire) F=flow ctl

## Use these for compatibility with stock ESIX (tty00h = dial-in + flow ctl):
#$I_fmt   = "%02d";
#$IF_fmt  = "%02dh";
#$O_fmt   = "%02da";	# 'a' to suggest "ACU" as in uucp Devices file.
#$OF_fmt  = "%02dah";
#$D_fmt   = "%02dd";	# 'd' for direct, no modem control (3-wire)

# Use these for names similar to those suggested by README.FAS (tty[F][M]##):
$I_fmt   = "M%02d";	# 'M' for wait-for-Modem-carrier (for getty)
$IF_fmt  = "FM%02d";	# 'F' adds hardware flow control
$O_fmt   = "%02d";
$OF_fmt  = "F%02d";
$D_fmt   = "D%02d";	# 'D' for direct, no modem control (3-wire)


# ---------- you should not have to change anything below here ----------

# Minor number bits OR'ed together with device number:

$I_minor  = 0xC0;	# Wait for carrier unless opened with O_NBLOCK
$O_minor  = 0x40;	# Ignore carrier until (and unless) it appears 
$D_minor  = 0;		# Direct without any modem control (3 wire)

$F_minor  = 0x10;	# full-duplex hardware flow control
$nF_minor = 0;		# NO full-duplex hardware flow control

$dvrname = "sas";
$outfile = "Node";
$outfile_sav = "$outfile.OLD";

if (-f $outfile && -f $outfile_sav) {
	print "$outfile_sav exists - can't save $outfile\n";
	print "Please remove $outfile or $outfile_sav and try again\n";
	exit(1);
}

if ($ARGV[0] ne "-d") {
    $buf = ""; # avoid -w warning
    print "How many serial ports [$nports] ?"; 
	($buf = <STDIN>) || die; chop $buf;
	($buf ne "") && ($nports = $buf);
	($nports < 1 || $nports > 16) && die "Invalid number of ports\n";
    print "Create devices with forced full-duplex hardware flow ctl [$flow] ?";
	($buf = <STDIN>) || die; chop $buf;
	($buf ne "") && ($flow = $buf);
    print "Create regular (non-hardware-flow-control) devices [$noflow] ?";
	($buf = <STDIN>) || die; chop $buf;
	($buf ne "") && ($noflow = $buf);
    print "Create direct-connect (without any modem control) devices [$direct]? ";
	($buf = <STDIN>) || die; chop $buf;
	($buf ne "") && ($direct = $buf);
    print "Create aliases in /dev/term/ directory [$termdir]? ";
	($buf = <STDIN>) || die; chop $buf;
	($buf ne "") && ($termdir = $buf);
}	

if ($termdir =~ /^[Yy]/) {
	@prefixlist = ('tty', 'term/');
} else {
	@prefixlist = ('tty');
}

if (-f $outfile) {
	rename($outfile, $outfile_sav) || die "Can't save $outfile_sav: $!\n";
	print "Existing \"$outfile\" saved as $outfile_sav\n";
}
open(OUT, ">$outfile") || die "Can't create $outfile: $!\n";
print STDOUT "Creating \"$outfile\" ...\n";

for ($i=0; $i < $nports; $i++) {
	print "\n";
	if ($flow =~ /^[Yy]/) {
		&do_output( sprintf($IF_fmt, $i), $I_minor | $F_minor | $i );
		&do_output( sprintf($OF_fmt, $i), $O_minor | $F_minor | $i );
	}
	if ($noflow =~ /^[Yy]/) {
		&do_output( sprintf($I_fmt, $i), $I_minor | $nF_minor | $i );
		&do_output( sprintf($O_fmt, $i), $O_minor | $nF_minor | $i );
	}
	if ($direct =~ /^[Yy]/) {
		&do_output( sprintf($D_fmt, $i), $D_minor | $i );
	}	
}

sub do_output {
	local ($name_suffix, $minor) = @_;
	local ($desc, $prefix, $name);

	# Form description of the minor bits we've been handed
	$desc = "";
	if    ((($minor >> 6) & 3) == 0) 
		{ $desc .= "direct (always ignore carrier)"; }
	elsif ((($minor >> 6) & 3) == 1) 
		{ $desc .= "dial-OUT (ignore CD initially)"; }
	elsif ((($minor >> 6) & 3) == 2)
		{ $desc .= "wait for carrier (O_NBLOCK not allowed)[not recommended]"; }
	else                          
		{ $desc .= "dial-in (wait for carrier)"; }
	if    ((($minor >> 4) & 3) == 0) 
		{ $desc .= ""; }
	elsif ((($minor >> 4) & 3) == 1) 
		{ $desc .= ", hardware flow ctl"; }
	elsif ((($minor >> 4) & 3) == 2) 
		{ $desc .= ", half-duplex output flow ctl [not recommended]"; }
	else 
		{ $desc .= ", half-duplex flow ctl, with buffer state signal"; }

	# Generate device files per @prefixlist, also printing descriptions
	foreach $prefix (@prefixlist) {
		$name = $prefix . $name_suffix;
		print OUT "$dvrname	$name	c	$minor\n";
		printf STDOUT "%-9s %3d (%s %s)  %s\n", 
			      $name, $minor, 
			      &fmt_binary(4, $minor>>4), 
			      &fmt_binary(4, $minor & 0xF), 
			      $desc;
		$desc = "";  # show description only on first of equiv lines
	}
}

sub fmt_binary {
	local ($minwidth, $val) = @_;
	local ($result) = "";
	while ($val != 0 || $minwidth > 0) {
		$result = pack("c", (ord("0") + ($val & 1))) . $result;
		$val >>= 1;
		$minwidth--;
	}
	return $result;
}
