/* 
** Copyright 1993 by Miron Livny, Mike Litzkow, and Emmanuel Ackaouy.
** 
** Permission to use, copy, modify, and distribute this software and its
** documentation for any purpose and without fee is hereby granted,
** provided that the above copyright notice appear in all copies and that
** both that copyright notice and this permission notice appear in
** supporting documentation, and that the names of the University of
** Wisconsin and the copyright holders not be used in advertising or
** publicity pertaining to distribution of the software without specific,
** written prior permission.  The University of Wisconsin and the 
** copyright holders make no representations about the suitability of this
** software for any purpose.  It is provided "as is" without express
** or implied warranty.
** 
** THE UNIVERSITY OF WISCONSIN AND THE COPYRIGHT HOLDERS DISCLAIM ALL
** WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES
** OF MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE UNIVERSITY OF
** WISCONSIN OR THE COPYRIGHT HOLDERS BE LIABLE FOR ANY SPECIAL, INDIRECT
** OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS
** OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE
** OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE
** OR PERFORMANCE OF THIS SOFTWARE.
** 
** Author:   Emmanuel Ackaouy
**
*/ 


#include <stdio.h>
#include <string.h>
#if defined(AIX31) || defined(AIX32)
#include <time.h>
#else /* AIX31 or AIX32 */
#include <sys/time.h>
#endif /* AIX31 or AIX32 */

#include <sys/types.h>
#include "condor_types.h"
#include "debug.h"
#include "expr.h"
#include "sched.h"
#include "stats.h"
#include <sys/socket.h>
#include <netinet/in.h>
#include "manager.h"


char *strdup();

extern MACH_REC	*MachineList;
extern char		*StatsFile;

/*
***		KEY_REC FUNCTIONS
 */


KEY_REC *create_key_rec()
{
	KEY_REC	*tmp;

	if ((tmp = (KEY_REC *) MALLOC(sizeof(KEY_REC))) == (KEY_REC *)0){
		dprintf(D_ALWAYS, "Out of memory\n");
		return (KEY_REC *) 0;
	}

	tmp->key = (char *) 0;
	tmp->count = 0;
	tmp->sum = 0;
	tmp->sum2 = 0;
	tmp->next = (KEY_REC *) 0;

	return tmp;
}


void delete_key_rec(ptr)
KEY_REC	*ptr;
{
	if (!ptr)
		return;

	FREE(ptr->key);
	FREE(ptr);
}


void display_key_rec(ostream, ptr, curr)
FILE	*ostream;
KEY_REC	*ptr;
int		curr;
{
	if (curr == -1){
		fprintf(ostream, "\"%s\": %d %d %.0f %d %d\n", ptr->key,
						ptr->count, ptr->sum, ptr->sum2, ptr->min, ptr->max);
	}
	else{
		fprintf(ostream, "\"%s\": [%d] %d %d %.0f", ptr->key, curr,
						ptr->count+1, ptr->sum + curr, (ptr->sum2 + curr*curr));
		if (ptr->count == 0){
			fprintf(ostream, " %d %d\n", curr, curr);
		}
		else{
			fprintf(ostream, " %d %d\n", (curr < ptr->min) ? curr : ptr->min,
										(curr > ptr->max) ? curr : ptr->max);
		}
	}
}



/*
***		KEY_ACC FUNCTIONS
 */


KEY_ACC *create_key_acc()
{
	KEY_ACC	*tmp;

	if ((tmp = (KEY_ACC *) MALLOC(sizeof(KEY_ACC))) == (KEY_ACC *)0){
		dprintf(D_ALWAYS, "Out of memory\n");
		return (KEY_ACC *) 0;
	}

	tmp->list = (KEY_REC *) 0;
	tmp->curr = (KEY_REC *) 0;

	return tmp;
}


void delete_key_acc(ptr)
KEY_ACC	*ptr;
{
	KEY_REC	*trav_ptr;

	if (!ptr)
		return;

	for(trav_ptr = ptr->list; trav_ptr; trav_ptr = trav_ptr->next)
		delete_key_rec(trav_ptr);
	FREE(ptr);
}


void add_key_acc(ptr, key, time_stamp)
KEY_ACC	*ptr;
char	*key;
int		time_stamp;
{
	char	*tmp_key;
	KEY_REC	*trav;
	KEY_REC	*prev;
	int		cmp_val;


		/* if key is null or empty -> record as unknown key */
	if (key == (char *) 0 || *key == '\0'){
		if ((tmp_key = (char *) MALLOC(strlen("(???)")+1)) == (char *)0){
			dprintf(D_ALWAYS, "Out of memory\n");
			return;
		}
		memcpy(tmp_key, "(???)", strlen("(???)")+1);
	}
	else{
		if ((tmp_key = (char *) MALLOC(strlen(key)+1)) == (char *)0){
			dprintf(D_ALWAYS, "Out of memory\n");
			return;
		}
		memcpy(tmp_key, key, strlen(key)+1);
	}

		/* if key has not changed, do not update stats */
	if (ptr->curr != (KEY_REC *)0 &&
		memcmp(ptr->curr->key, tmp_key, strlen(tmp_key)+1) == 0){

		FREE(tmp_key);
		return;
	}

		/* if the key has changed, update the stats for the last one */
	if (ptr->curr != (KEY_REC *)0){
		ptr->curr->sum += time_stamp - ptr->curr->start_time;
		ptr->curr->sum2 += (double) (time_stamp - ptr->curr->start_time) *
						   (double) (time_stamp - ptr->curr->start_time);
		if (++ ptr->curr->count == 1){
			ptr->curr->min = time_stamp - ptr->curr->start_time;
			ptr->curr->max = time_stamp - ptr->curr->start_time;
		}
		else{
			if ((time_stamp - ptr->curr->start_time) < ptr->curr->min)
				ptr->curr->min = time_stamp - ptr->curr->start_time;
			if ((time_stamp - ptr->curr->start_time) > ptr->curr->max)
				ptr->curr->max = time_stamp - ptr->curr->start_time;
		}
	}

		/* set up the new key */
	for(trav = ptr->list, prev = (KEY_REC *)0; trav;){
		cmp_val = memcmp(tmp_key, trav->key, strlen(tmp_key)+1);

		if (cmp_val > 0){				/* continue unless last */
			prev = trav;
			if ((trav = trav->next) == (KEY_REC *)0){
				prev->next = create_key_rec();
				prev->next->key = tmp_key;
				prev->next->start_time = time_stamp;
				ptr->curr = prev->next;

				return;
			}
		}
		else if (cmp_val == 0){			/* match found */
			FREE(tmp_key);
			trav->start_time = time_stamp;
			ptr->curr = trav;

			return;
		}
		else{							/* insert new just before */
			if (prev == (KEY_REC *)0){		/* start of list */
				ptr->list = create_key_rec();
				ptr->list->key = tmp_key;
				ptr->list->start_time = time_stamp;
				ptr->list->next = trav;
				ptr->curr = ptr->list;
			}
			else{							/* current pos */
				prev->next = create_key_rec();
				prev->next->key = tmp_key;
				prev->next->start_time = time_stamp;
				prev->next->next = trav;
				ptr->curr = prev->next;
			}

			return;
		}
	}

		/* list was null */
	ptr->list = create_key_rec();
	ptr->list->key = tmp_key;
	ptr->list->start_time = time_stamp;
	ptr->curr = ptr->list;

	return;
}


void display_key_acc(ostream, ptr, the_time)
FILE	*ostream;
KEY_ACC	*ptr;
int		the_time;
{
	KEY_REC	*trav;
	int		tmp_sum;

	for(trav = ptr->list; trav; trav = trav->next){
		if (trav == ptr->curr){		/* print time in current state */
			display_key_rec(ostream, trav, the_time - trav->start_time);
		}
		else{						/* not the current state */
			display_key_rec(ostream, trav, -1);
		}
	}
}



/*
***		MACH_STATS FUNCTIONS
 */


MACH_STATS *create_mach_stats()
{
	MACH_STATS	*tmp;

	if ((tmp = (MACH_STATS *) MALLOC(sizeof(MACH_STATS))) == (MACH_STATS *)0){
		dprintf(D_ALWAYS, "Out of memory\n");
		return (MACH_STATS *) 0;
	}

	tmp->state = create_key_acc();

	return tmp;
}


void delete_mach_stats(ptr)
MACH_STATS	*ptr;
{
	if (!ptr)
		return;

	delete_key_acc(ptr->state);
	FREE(ptr);
}


void reset_mach_stats(ptr)
MACH_STATS	*ptr;
{
	if (!ptr)
		return;

	delete_key_acc(ptr->state);
	ptr->state = create_key_acc();
}


void display_mach_stats(ostream, ptr, the_time)
FILE		*ostream;
MACH_STATS	*ptr;
int			the_time;
{
	if (!ptr)
		return;
	
		/* state */
	display_key_acc(ostream, ptr->state, the_time);
}


int avail(ptr)
MACH_REC	*ptr;
{
	float	ldavg;
	int		kbdidle;

	if (evaluate_int("KeyboardIdle", &kbdidle, ptr->machine_context,
										(CONTEXT *) 0) < 0){
		return FALSE;	/* assume not available */
	}
	if (evaluate_float("LoadAvg", &ldavg, ptr->machine_context,
										(CONTEXT *) 0) < 0){
		return FALSE;	/* assume not available */
	}

	if (ldavg < .3  &&  kbdidle > 15*60) return TRUE;

	return FALSE;
}


/*
***		INTERFACE FUNCTIONS
***	(The create and delete functions are listed in the above groups)
 */

void update_mach_stats(ptr, kind, time_stamp)
MACH_REC	*ptr;
int			kind;
int			time_stamp;
{
	char	*tmpstate;

	switch(kind){
		case STARTD_STATS:
			if (strcmp(ptr->line->state, "Running") == MATCH){
				tmpstate = strdup("Condor");
			}
			else if (strcmp(ptr->line->state, "NoJob") == MATCH){
				if (avail(ptr)) tmpstate = strdup("Avail");
				else tmpstate = strdup("Owner");
			}
			else tmpstate = strdup(ptr->line->state);

			add_key_acc(ptr->stats->state, tmpstate, time_stamp);
			free(tmpstate);

			break;

		case SCHEDD_STATS:
			break;
	}
}


void display_stats()
{
	FILE		*ostream;
	MACH_REC	*trav;
	int		the_time;

	if ((ostream = fopen(StatsFile, "a")) == NULL){
		dprintf(D_ALWAYS, "Could not open '%s' to dump stats\n", StatsFile);
		return;
	}

	the_time = time((time_t *)0);

	fprintf(ostream, "*** %d\n", the_time);

	for(trav = MachineList->next; trav->name; trav = trav->next){
		fprintf(ostream, "\"%s\":\n", trav->name);
		display_mach_stats(ostream, trav->stats, the_time);
		fprintf(ostream, "--\n");
	}

	fclose(ostream);
	dprintf(D_ALWAYS, "DUMPed statistics in '%s'\n", StatsFile);
}
