/* 
** Copyright 1993 by Miron Livny, and Mike Litzkow
** 
** Permission to use, copy, modify, and distribute this software and its
** documentation for any purpose and without fee is hereby granted,
** provided that the above copyright notice appear in all copies and that
** both that copyright notice and this permission notice appear in
** supporting documentation, and that the names of the University of
** Wisconsin and the copyright holders not be used in advertising or
** publicity pertaining to distribution of the software without specific,
** written prior permission.  The University of Wisconsin and the 
** copyright holders make no representations about the suitability of this
** software for any purpose.  It is provided "as is" without express
** or implied warranty.
** 
** THE UNIVERSITY OF WISCONSIN AND THE COPYRIGHT HOLDERS DISCLAIM ALL
** WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES
** OF MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE UNIVERSITY OF
** WISCONSIN OR THE COPYRIGHT HOLDERS BE LIABLE FOR ANY SPECIAL, INDIRECT
** OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS
** OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE
** OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE
** OR PERFORMANCE OF THIS SOFTWARE.
** 
** Author:  Mike Litzkow
**
*/ 

#define _POSIX_SOURCE

#include "condor_common.h"
#include "condor_config.h"
#include "condor_status.h"
#include "condor_constants.h"
#include "condor_debug.h"
#include "condor_jobqueue.h"
#include "condor_network.h"
#include "condor_string.h"
#include "proc_obj.h"
#include "alloc.h"
#include "sched.h"

static char *_FileName_ = __FILE__;		/* Used by EXCEPT (see except.h)     */


DBM		*Q;

char	*MyName;
char	*Spool;
char *CollectorHost;

int Long = 0;
int Prioritized = 0;
char *MatchingArch_Os = NULL;
char *MatchingUser = NULL;
char *MatchingHost = NULL;

#if defined(OSF1)
#pragma define_template List<char>
#pragma define_template Item<char>
#endif

List<char> *MachinesWithJobs;
List<char> *MachinesOfInterest;
List<char> *ArchitectureOfInterest;
List<char> *OpsysOfInterest;
List<char> *UsersOfInterest;

void init_params();
void build_machine_list();
void usage();
BOOLEAN interested( const char *, List<char> * );
void do_display( ProcObj *p, char *host );
extern "C" BOOLEAN xdr_proc( XDR *, GENERIC_PROC * );
extern "C" BOOLEAN snd_int(XDR *, int, BOOLEAN);
void process_machine( char *hostname );

/*
  Tell folks how to use this program.
*/
void
usage()
{
	fprintf( stderr, 
		"Usage: %s [-l] [-p] [-a arch] [-o opsys] [-u user] [-h hostname]\n", 
			MyName );
	exit( 1 );
}

int
main( int argc, char *argv[] )
{
	int		i;
	char	*arg;
	char	*m;

	MyName = argv[0];
	MachinesOfInterest = new List<char>;
	UsersOfInterest = new List<char>;
	ArchitectureOfInterest = new List<char>;
	OpsysOfInterest = new List<char>;

	for( argv++; arg = *argv; argv++ ) {
		if( arg[0] == '-' && arg[1] == 'p' ) {
			Prioritized++;
		} 
		else if( arg[0] == '-' && arg[1] == 'l' ) {
			Long++;
		} 
		else if( arg[0] == '-' && arg[1] == 'o' ) {
			arg = *(++argv);
			if( arg ) {
				lower_case( arg );
				OpsysOfInterest->Append(arg);
			} else {
				usage();
			}
		} 
		else if( arg[0] == '-' && arg[1] == 'a' ) {
			arg = *(++argv);
			if( arg ) {
				lower_case( arg );
				ArchitectureOfInterest->Append(arg);
			} else {
				usage();
			}
		} 
		else if( arg[0] == '-' && arg[1] == 'u' ) {
			arg = *(++argv);
			if( arg ) {
				UsersOfInterest->Append(arg);
			} else {
				usage();
			}
		} 
		else if( arg[0] == '-' && arg[1] == 'h' ) {
			arg = *(++argv);
			if( arg ) {
				MachinesOfInterest->Append( arg );
			} else {
				usage();
			}
		} 
		else if( arg[0] == '-') {
			usage();
		}
	}

	config( MyName, (CONTEXT *)0 );
	init_params();

	MachinesWithJobs = new List<char>;
	build_machine_list();

		// process each machine
	MachinesWithJobs->Rewind();
	while( m = MachinesWithJobs->Next() ) {
		if( interested(m,MachinesOfInterest) ) {
			process_machine( m );
		}
	}
	if  (!MachinesOfInterest->IsEmpty())  
	{
		MachinesOfInterest->Rewind();
		while ( m = MachinesOfInterest->Next() )
			process_machine( m );
	}

	MachinesWithJobs->Rewind();
	while( m = MachinesWithJobs->Next() ) {
		delete [] m;
	}
	delete MachinesWithJobs;

	delete MachinesOfInterest;
	delete UsersOfInterest;
	delete ArchitectureOfInterest;
	delete OpsysOfInterest;

#if defined(ALLOC_DEBUG)
	print_alloc_stats();
#endif
	return 0;

}

void
build_machine_list()
{
	int			sock = -1;
	XDR			xdr, *xdrs = NULL;
	int			i;
	int			cmd;
	STATUS_LINE	*line;


		// Connect to the collector
	if( (sock = do_connect(CollectorHost, "condor_collector",
													COLLECTOR_PORT)) < 0 ) {
		dprintf( D_ALWAYS, "Can't connect to Condor Collector\n" );
		exit( 1 );
	}
	xdrs = xdr_Init( &sock, &xdr );
	xdrs->x_op = XDR_ENCODE;

	cmd = GIVE_STATUS_LINES;
	ASSERT( xdr_int(xdrs, &cmd) );
	ASSERT( xdrrec_endofrecord(xdrs,TRUE) );

	xdrs->x_op = XDR_DECODE;

	for(;;) {
		line = new STATUS_LINE;
		memset( line, 0, sizeof(STATUS_LINE) );
		if( !xdr_status_line( xdrs, line ) ) {
			EXCEPT( "Can't read status line from Condor Collector" );
		}
		if( line->name == NULL || line->name[0] == '\0' ) {
			break;
		}
		if( line->tot != 0 ) {
			MachinesWithJobs->Append( string_copy(line->name) );
		}
	}
	xdr_destroy( xdrs );
	(void) close( sock );
}


BOOLEAN
interested( const char *this_one, List<char> *list )
{
	char	*cur;

	if( list->IsEmpty() ) {
		return TRUE;
	}

	list->Rewind();
	while( cur = list->Next() ) {
		if( strcmp(cur,this_one) == MATCH ) {
			return TRUE;
		}
	}
	return FALSE;
}


/*
  Get and display job queue for a single machine.
*/
void
process_machine( char *hostname )
{
	int		schedSocket;
	int		cmd;
	XDR		xdr, *xdrs;
	ProcObj			*p;


	if( Prioritized ) {
		cmd = SEND_ALL_JOBS_PRIO;
	} else {
		cmd = SEND_ALL_JOBS;
	}

	if( (schedSocket = do_connect( hostname, "condor_schedd", 
												SCHED_PORT)) < 0 ) {
		fprintf( stderr, "Can't connect to schedd on %s\n", hostname);
		return;
	}

	xdrs = xdr_Init( &schedSocket, &xdr );

	if( !snd_int(xdrs, cmd, TRUE) ) {
		dprintf( D_ALWAYS, 
				"Can't send command to schedd on %s\n", hostname );
		xdr_destroy( xdrs );
		(void)close( schedSocket );
	}

	for( p = ProcObj::create(xdrs); p; p = ProcObj::create(xdrs) ) {
		if(
			interested( p->get_owner(), UsersOfInterest ) &&
			interested( p->get_arch(), ArchitectureOfInterest ) &&
			interested( p->get_opsys(), OpsysOfInterest )) {
				do_display( p, hostname );
		}
		delete p;
	}

	xdr_destroy( xdrs );
	(void)close( schedSocket );
}

void
do_display( ProcObj	*p, char *host )
{
	static char *last_host = "";

	if( strcmp(host,last_host) != MATCH ) {
		printf( "\n%s:\n", host);
		last_host = host;
		ProcObj::special_header();
	}
	
	if( Long ) {
		p->display();
	} else {
		p->display_special();
	}

}

void
init_params()
{
	if( (CollectorHost = param("COLLECTOR_HOST")) == NULL ) {
		EXCEPT( "COLLECTOR_HOST not specified in config file\n" );
	}
}

extern "C" int SetSyscalls( int foo ) { return foo; }
